
// ************************ BEGIN_COPYRIGHT_NOTICE ************************
//
// This source code copyright (c) 1995 by Paul Keister and Eric Dill
//
// Permission to use, copy, modify, distribute, and sell this software
// and its documentation for any purpose is hereby granted without
// fee, provided that this copyright notice appear in all copies
// and that the author's name not be used in advertising or publicity
// pertaining to distribution of the software without specific, written
// prior permission.  The author makes no representations about the
// suitability of this software for any purpose.  It is provided "as
// is" without express or implied warranty.
//
// Please send all bug fixes, ideas, and comments to:
//
//                                      CompuServe 71242,1145 / 75061,3310
//
// ************************* END_COPYRIGHT_NOTICE *************************

#include "prtStream.h"


// ****************************************************
// *                                                  *
// * NOTE:  ...::Render functions _must_ extract all  *
// *        but the terminating ESC from the istream& *
// *                                                  *
// ****************************************************


//
// Draw a Single Box
//

void SpecialEffectManip< Box >
   ::Render( prtStream *pstream, istream& in )
{
   float top, bottom, left, right;

   in >> top >> bottom  >> left >> right;

   top    *= pstream->yScale();
   bottom *= pstream->yScale();
   left   *= pstream->xScale();
   right  *= pstream->xScale();

   pstream->PDC().SelectStockObject( BLACK_PEN );

   pstream->PDC().MoveTo( left,  top );
   pstream->PDC().LineTo( right, top );
   pstream->PDC().LineTo( right, bottom );
   pstream->PDC().LineTo( left,  bottom );
   pstream->PDC().LineTo( left,  top );
}


//
// Draw a Frame
//

void SpecialEffectManip< Frame >
   ::Render( prtStream *pstream, istream& in )
{
   float top, left,  width, height, lineWidth;
   int   lineStyle,  red,   green,  blue;


   in >> top       >> left      >> width >> height
      >> lineWidth >> lineStyle >> red   >> green  >> blue;

   top       *= pstream->yScale();
   left      *= pstream->xScale();
   width     *= pstream->xScale();
   height    *= pstream->yScale();
   lineWidth *= pstream->xScale();

   float bottom = top  + height;
   float right  = left + width;

   if ( lineStyle != PS_SOLID )
      lineWidth = 1;

   TPen  pen( TColor( red, green, blue ), lineWidth, lineStyle );
   pstream->PDC().SelectObject( pen );

   pstream->PDC().MoveTo( left,  top );
   pstream->PDC().LineTo( right, top );
   pstream->PDC().LineTo( right, bottom );
   pstream->PDC().LineTo( left,  bottom );
   pstream->PDC().LineTo( left,  top );

   pstream->PDC().RestorePen();
}


//
// Draw a Bitmap
//

void SpecialEffectManip< BitMap >
      ::Render( prtStream *pstream, istream& in )
{
   //
   // read in bitmap position & size
   //

   float xPos, yPos, width, height;

   in >> xPos >> yPos >> width >> height >> ws;

   xPos   *= pstream->xScale();
   yPos   *= pstream->yScale();
   height *= pstream->yScale();
   width  *= pstream->xScale();


   //
   // read in bitmap resource ID
   //

   const resSIZE = 32;
   char  resName[resSIZE];

   in.getline( resName, resSIZE, 0 );  // get bitmap name
   resName[in.gcount()] = 0;           // terminate the bitmap name

   TResId ResId( resName );


   //
   // process the bitmap
   //

   if( !( pstream->PDC().GetDeviceCaps( RASTERCAPS ) & RC_BITBLT ) )
      return;  // if printer doesn't support bitmaps

   if ( !xPos )
      xPos = pstream->LeftMargin( 0 ); // left margin in unscaled units

   if ( !yPos )
      yPos = pstream->yPrintPos();

   TPoint bmpPos(  xPos,   yPos   );
   TPoint optSize( width,  height );
   TDib   bm(     *Module, ResId  );

   pstream->PDC().LPtoDP( &bmpPos,  1 );
   pstream->PDC().LPtoDP( &optSize, 1 );

   int   oldMode = pstream->PDC().SetMapMode( MM_TEXT );

   TSize bmSize( bm.Width(),   bm.Height() );
   TRect src(    TPoint(0, 0), bmSize );
   TRect dest;
   if ( width + height )
      dest = TRect( bmpPos, TSize(optSize) );
   else
      dest = TRect( bmpPos, bmSize );

   pstream->PDC().StretchDIBits( dest, src, bm );
   pstream->PDC().SetMapMode( oldMode );
}


//
// Set the Font
//

void SpecialEffectManip< Font >
   ::Render( prtStream *pstream, istream& in )
{
   LOGFONT  lf;

   in >> lf.lfHeight          >> lf.lfWidth
      >> lf.lfEscapement      >> lf.lfOrientation
      >> lf.lfWeight          >> lf.lfItalic
      >> lf.lfUnderline       >> lf.lfStrikeOut
      >> lf.lfCharSet         >> lf.lfOutPrecision
      >> lf.lfClipPrecision   >> lf.lfQuality
      >> lf.lfPitchAndFamily;

   in.getline( lf.lfFaceName, LF_FACESIZE, 0 );    // get face name
   lf.lfFaceName[in.gcount()] = 0;       // terminate the face name

   if ( lf.lfHeight > 0 )
      lf.lfHeight = -MulDiv( lf.lfHeight, abs(pstream->yScale()), 72 );

   pstream->SetFont( lf );
}


//
// Set the Left Margin
//

void SpecialEffectManip< LeftMargin >
   ::Render( prtStream *pstream, istream& in )
{
   float margin;  // left margin (inches from left)

   in >> margin;

   pstream->LeftMargin( margin );
}


//
// Set the Top Margin
//

void SpecialEffectManip< TopMargin >
   ::Render( prtStream *pstream, istream& in )
{
   float margin;  // top margin (inches from top)

   in >> margin;

   pstream->TopMargin( margin );
}


//
// Set the Print Pos
//

void SpecialEffectManip< SetPrintPos >
   ::Render( prtStream *pstream, istream& in )
{
   float yPos, xPos;
   bool  ignoreMargins;

   in >> yPos >> xPos >> ignoreMargins;

   pstream->SetPrintPos( yPos, xPos, ignoreMargins );
}


//
// Save a prtStream state value
//

void SpecialEffectManip< SaveState >
   ::Render( prtStream *pstream, istream& in )
{
   int      state;
   int      index;

   in >> state >> index;

   pstream->SaveState( prtStreamState(state), index );
}


//
// Restore a prtStream state value
//

void SpecialEffectManip< RestoreState >
   ::Render( prtStream *pstream, istream& in )
{
   int      state;
   int      index;

   in >> state >> index;

   pstream->RestoreState( prtStreamState(state), index );
}


//
// Page Break
//

#pragma argsused
void SpecialEffectManip< PageBreak >
   ::Render( prtStream *pstream, istream& in )
{
   in >> ws;      // skip trailing WhiteSpace
   pstream->NewPage();
}


//
// Conditional Page Break
//

void SpecialEffectManip< CondPageBreak >
   ::Render( prtStream *pstream, istream& in )
{
   float printSize,
         pageLimit;

   in >> printSize >> pageLimit;

   printSize *= pstream->yScale();
   pageLimit *= pstream->yScale();

   if ( abs( pstream->yPrintPos() + printSize ) > abs( pageLimit ) )
      pstream->NewPage();
}


//
// Page Number
//

void SpecialEffectManip< PrtPageNum >
   ::Render( prtStream *pstream, istream& in )
{
   char     buff[8] = {0,0,0,0,0,0,0,0};

   pstream->PrintLine( itoa( pstream->PageNum(), buff, 10 ) );

   in >> ws;
}


//
// Process a Page Header specification
//

void SpecialEffectManip< StartHeader >
      ::Render( prtStream *pstream, istream& in )
{
   pstream->SaveState();

   const char *cursor = pstream->Cursor();

   while ( *(++cursor) && *cursor != ESC )
      ;

   pstream->HeaderStart( cursor + 1 );

   in >> ws;      // skip trailing whitespace
}


void SpecialEffectManip< StartFooter >
      ::Render( prtStream *pstream, istream& in )
{
   pstream->SaveState();

   const char *cursor = pstream->Cursor();

   while ( *(++cursor) && *cursor != ESC )
      ;

   pstream->FooterStart( cursor + 1 );

   in >> ws;      // skip trailing whitespace
}


void SpecialEffectManip< EndHeader>
      ::Render( prtStream *pstream, istream &in )
{
   pstream->RestoreState();

   const char *cursor = pstream->Cursor();

   while ( *(--cursor) && *cursor != ESC )
      ;

   pstream->HeaderEnd( cursor - 1 );

   in >> ws;   // skip trailing whitespace
}


void SpecialEffectManip< EndFooter>
      ::Render( prtStream *pstream, istream &in )
{
   pstream->RestoreState();

   const char *cursor = pstream->Cursor();

   while ( *(--cursor) && *cursor != ESC )
      ;

   pstream->FooterEnd( cursor - 1 );

   in >> ws;   // skip trailing whitespace
}

