
// ************************ BEGIN_COPYRIGHT_NOTICE ************************
//
// This source code copyright (c) 1995 by Paul Keister and Eric Dill
//
// Permission to use, copy, modify, distribute, and sell this software
// and its documentation for any purpose is hereby granted without
// fee, provided that this copyright notice appear in all copies
// and that the author's name not be used in advertising or publicity
// pertaining to distribution of the software without specific, written
// prior permission.  The author makes no representations about the
// suitability of this software for any purpose.  It is provided "as
// is" without express or implied warranty.
//
// Please send all bug fixes, ideas, and comments to:
//
//                                      CompuServe 71242,1145 / 75061,3310
//
// ************************* END_COPYRIGHT_NOTICE *************************


// ************************ BEGIN_VERSION_HISTORY *************************
//
// 01-May-95 - V 1.00 - Posted on CompuServe
//
// ************************* END_VERSION_HISTORY **************************

#include "prtStream.h"


//
// prtStream constructor
//

prtStream::prtStream( float lMargin, float tMargin )
   : initLeftMargin(lMargin), initTopMargin(tMargin), ostrstream()
{
   //
   // initialize prtStream data members
   //

   Parent      = 0;
   Font        = 0;
   xSCALE      = ySCALE    = 1;
   curLine     = cursor    = tooFar      = 0;
   headerStart = headerEnd = footerStart = footerEnd = 0;
   newPage     = newLine   = true;
   leftMargin  = topMargin = printPos    = 0;
   lineLen     = pageNum   = 0;
   pdc         = 0;


   //
   // initialize function array to null value
   //

   for( int counter = 0; counter < LastEffect; counter++ )
      RenderFunctions[ counter ] = 0;
}


//
// prtStream destructor
// this no longer prints orphan objects
//

prtStream::~prtStream( void )
{
   delete Font;
}


//
// prtStream::ConfigureMapMode makes all MapMode adjustments
//

void prtStream::ConfigureMapMode( void )
{
   pdc->SetMapMode(   MM_HIENGLISH );     // one logical unit = 0.001 inch
   pdc->SetTextAlign( TA_UPDATECP  );     // update position after text output
   pdc->GetTextMetrics( tm );             // get text metrics
   xSCALE = 1000;
   ySCALE = 1000;


   switch ( pdc->GetMapMode() )
      {
         case MM_TEXT:
            Parent->MessageBox( "MapMode = MM_TEXT", "Warning from prtStream::ConfigureMapMode()" );
            xSCALE = pdc->GetDeviceCaps( LOGPIXELSX );
            ySCALE = pdc->GetDeviceCaps( LOGPIXELSY );
            break;

         case MM_ISOTROPIC:
            Parent->MessageBox( "MapMode = MM_ISOTROPIC", "*** Error in prtStream::ConfigureMapMode() ***" );
            break;

         case MM_ANISOTROPIC:
            Parent->MessageBox( "MapMode = MM_ANISOTROPIC", "*** Error in prtStream::ConfigureMapMode() ***" );
            break;

         case MM_HIENGLISH:
         case MM_LOENGLISH:
         case MM_HIMETRIC:
         case MM_LOMETRIC:
         case MM_TWIPS:
            ySCALE      *= -1;
            tm.tmHeight *= -1;
            break;

         default:
            Parent->MessageBox( "Unknown MapMode", "*** Error in prtStream::ConfigureMapMode() ***" );
            break;
      }
}


//
// prtStream::PrintLine prints the current line
//

void
prtStream::PrintLine( void )
{
   if ( lineLen )
      {
         if ( newLine )
            pdc->MoveTo( leftMargin, printPos );

         pdc->TextOut( 0, 0, curLine, lineLen );
      }
}

void
prtStream::PrintLine( const char *buffer )   // print the buffer
{
   curLine = buffer;
   lineLen = strlen( buffer );
   newLine = false;
   PrintLine();
}

//
// prtStream::NewPage initiates a new page
//

void
prtStream::NewPage( void )
{
   PrintLine();
   newPage  =  true;
}


//
// prtStream::SetPrintPos resets the current print position
//
      
void
prtStream::SetPrintPos( float y, float x, bool ignoreMargin )
{
   int      xPos;
   TPoint   pos;

   pdc->GetCurrentPosition( pos );

   if ( x == 0 )
      x = pos.x / xScale();

   if ( y == 0 )
      y = pos.y / yScale();


   if ( x < LeftMargin() && !ignoreMargin )
      xPos = leftMargin;
   else
      xPos = x * xScale();

   if ( y < TopMargin() && !ignoreMargin )
      printPos = topMargin;
   else
      printPos = y * yScale();

   pdc->MoveTo( xPos, printPos );
   newLine = false;
}


//
// prtStream::SaveState saves a prtStream state value
//

void
prtStream::SaveState( prtStreamState state, int index )
{
   if ( index < 1 || index > MaxIndex )
      index = 0;

   int   value;

   switch( state )
      {
         case LeftMargin_: value   = leftMargin;
                           break;

         case TopMargin_:  value   = topMargin;
                           break;

         case PrintPos_:   value   = printPos;
                           break;

         default:          value   = 0;
                           break;
      }

   savedPos[ index ] = value;
}


//
// prtStream::RestoreState restores a prtStream state value
//

void
prtStream::RestoreState( prtStreamState state, int index )
{
   if ( index < 1 || index > MaxIndex )
      index = 0;

   int   value = savedPos[ index ];

   switch( state )
      {
         case LeftMargin_: leftMargin  = value;
                           break;

         case TopMargin_:  topMargin   = value;
                           break;

         case PrintPos_:   printPos    = value;
                           newLine = true;
                           break;

         default:          // do nothing
                           break;
      }
}


//
// prtStream::SetFont creates a new font
//

void
prtStream::SetFont( LOGFONT& lf )
{
   if ( Font )
      {
         pdc->RestoreFont();
         delete Font;
      }

   Font = new TFont( &lf );
   pdc->SelectObject( *Font );

   int direction = tm.tmHeight < 0 ? -1 : 1;

   pdc->GetTextMetrics( tm );
   tm.tmHeight *= direction;
}


//
// prtStream::SpecialEffect processes a special effect
//

void
prtStream::SpecialEffect( void )
{
   PrintLine();               // print any text collected to this point

   cursor += 2;               // skip past the special effect header

                              // create an input stream formatter
   istrstream  in( const_cast<char *>(cursor),
                   static_cast<int>( tooFar - cursor ) );
   int         eff;

   in >> eff;                 // read the special effect function index

   if ( eff >= 0 && eff < LastEffect && in.rdstate() == ios::goodbit )
      if ( RenderFunctions[ eff ] )
         (*RenderFunctions[ eff ])( this, in );

   cursor += static_cast<uint16>( in.tellg() );

   cursor++;                  // skip cursor past special effect

   curLine  = cursor;         // set start of new line
   lineLen  = 0;              // reset line length
}


//
// prtStreamPage class interfaces with Borland TPrintout class
//

class prtStreamPage : public TPrintout
{
   private:
      prtStream& Prt;

   public:
      prtStreamPage( prtStream& prt )
         :TPrintout( "prtStream print job" ), Prt( prt ) {}

      virtual void PrintPage( int page, TRect& rect, unsigned flags );
      virtual bool HasPage( int pageNumber )
         {
            return Prt.Cursor() < Prt.TooFar();
         }
};


//
// prtStream::Print starts the print process
//

void
prtStream::Print( TWindow *currentWin, TPrinter& printer )
{
   char *buffBase = str();

   Parent = currentWin;
   cursor = const_cast<const char *>(buffBase);
   tooFar = cursor + pcount();
   printer.Print( currentWin, prtStreamPage( *this ), false );
   delete   buffBase;
   cursor = 0;
}


//
// prtStreamPage::PrintPage is the TPrintout interface
//

void
prtStreamPage::PrintPage( int page, TRect&, unsigned )
{
   Prt.pdc = DC;
   Prt.PrintPage( page );
}


//
// prtStream::PrintPage prints a page
//

void
prtStream::PrintPage( int page )
{
   ConfigureMapMode();

   if ( page == 1 )
      {
         leftMargin = initLeftMargin * xSCALE;
         topMargin  = initTopMargin  * ySCALE;
      }

   pdc->MoveTo( leftMargin, topMargin );  // set start position

   printPos = topMargin;                  // printPos decreases from top of page
   lineLen  = 0;
   pageNum  = page;

   newPage = false;
   curLine = cursor;

   Scan( headerStart, headerEnd );

   while( cursor < tooFar && newPage == false )
      {
         ReadCursor();
      }

   Scan( footerStart, footerEnd );
}


void prtStream::Scan( const char *start, const char *pastEnd )
{
   int         safePos     = printPos; // save current print pos
   const char *safeCursor  = cursor;

   if( start )
      {
         cursor = start;

         if( !pastEnd )
            pastEnd = tooFar;

         while( cursor < pastEnd )
            ReadCursor();
      }

   cursor  = safeCursor;
   curLine = safeCursor;

   printPos = safePos;           // restore saved print pos
   newLine  = true;
}


void prtStream::ReadCursor( void )
{
   //
   // two escape characters signal a special effect
   //

   if( *cursor != ESC || *(cursor + 1) != ESC )
      {
         if ( *cursor == '\n' )
            {
               PrintLine();
               curLine      = ++cursor;
               lineLen      = 0;
               printPos    += tm.tmHeight;
               newLine      = true;
            }
         else
            {
               cursor++;
               lineLen++;
            }
      }
   else
      {
         SpecialEffect();
      }
}

